from mpmath import *


class mpdigits:
    def __init__(self, x=0):
        self.x =  x

    def get_d(self, d=pi):
        return mpf(d)

    def sin(self):
        return sin

    def cos(self):
        return cos

    def tan(self):
        return tan

    def sinh(self):
        return sinh

    def cosh(self):
        return cosh

    def tanh(self):
        return tanh

    def exp(self):
        return exp

    def sqrt(self):
        return sqrt

    def quad(self):
        return quad



from mpmath import *


class mptaylor:

    def __init__(self, f=sin, p=0, d=5):
        mp.dps = 15
        mp.pretty = True

        self.poly = taylor(f, p, d)
        nprint(chop(self.poly))


    def __len__(self):
        return len(self.poly)

    def __getitem__(self, idx):
        return self.poly[idx]

    def __setitem__(self, idx, val):
        self.poly[idx] = val


    def predict(self, x):
        return polyval(self.poly[::-1], x - p)



# BSD 3-Clause License
#
# Copyright (c) 2021, Dustin Kenefake
# All rights reserved.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#
# 1. Redistributions of source code must retain the above copyright notice, this
#    list of conditions and the following disclaimer.
#
# 2. Redistributions in binary form must reproduce the above copyright notice,
#    this list of conditions and the following disclaimer in the documentation
#    and/or other materials provided with the distribution.
#
# 3. Neither the name of the copyright holder nor the names of its
#    contributors may be used to endorse or promote products derived from
#    this software without specific prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
# DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE
# FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
# DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
# SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
# CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
# OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
# OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.




from mpmath import mp
import numpy



def bisection_search(f, low:float, high:float):
    """
    A root finding method that does not rely on derivatives

    :param f: a function f: X -> R
    :param low: the lower bracket
    :param high: the upper limit bracket
    :return: the location of the root, e.g. f(mid) ~ 0
    """
    # flip high and low if out of order
    if f(high) < f(low):
        low, high = high, low

    # find mid point
    mid = .5 * (low + high)

    while True:

        # bracket up
        if f(mid) < 0:
            low = mid
        # braket down
        else:
            high = mid

        # update mid point
        mid = .5 * (high + low)

        # break if condition met
        if abs(high - low) < 10 ** (-(mp.dps / 2)):
            break

    return mid


def concave_max(f, low:float, high:float):
    """
    Forms a lambda for the approximate derivative and finds the root

    :param f: a function f: X -> R
    :param low: the lower bracket
    :param high: the upper limit bracket
    :return: the location of the root f'(mid) ~ 0
    """
    # create an approximate derivative expression
    scale = high - low

    h = mp.mpf('0.' + ''.join(['0' for i in range(int(mp.dps / 1.5))]) + '1') * scale
    df = lambda x: (f(x + h) - f(x - h)) / (2.0 * h)

    return bisection_search(df, low, high)

def chev_points(n:int, lower:float = -1, upper:float = 1):
    """
    Generates a set of chebychev points spaced in the range [lower, upper]
    :param n: number of points
    :param lower: lower limit
    :param upper: upper limit
    :return: a list of multipressison chebychev points that are in the range [lower, upper]
    """
    #generate chebeshev points on a range [-1, 1]
    index = numpy.arange(1, n+1)
    range_ = abs(upper - lower)
    return [(.5*(mp.cos((2*i-1)/(2*n)*mp.pi)+1))*range_ + lower for i in index]


def remez(func, n_degree:int, lower:float=-1, upper:float=1, max_iter:int = 10):
    """
    :param func: a function (or lambda) f: X -> R
    :param n_degree: the degree of the polynomial to approximate the function f
    :param lower: lower range of the approximation
    :param upper: upper range of the approximation
    :return: the polynomial coefficients, and an approximate maximum error associated with this approximation
    """
    # initialize the node points

    x_points = chev_points(n_degree + 2, lower, upper)

    A = mp.matrix(n_degree + 2)
    coeffs = numpy.zeros(n_degree + 2)

    # place in the E column
    mean_error = float('inf')

    for i in range(n_degree + 2):
        A[i, n_degree + 1] = (-1) ** (i + 1)

    for i in range(max_iter):

        # build the system
        vander = numpy.polynomial.chebyshev.chebvander(x_points, n_degree)

        for i in range(n_degree + 2):
            for j in range(n_degree + 1):
                A[i, j] = vander[i, j]

        b = mp.matrix([func(x) for x in x_points])
        l = mp.lu_solve(A, b)

        coeffs = l[:-1]

        # build the residual expression
        r_i = lambda x: (func(x) - numpy.polynomial.chebyshev.chebval(x, coeffs))

        interval_list = list(zip(x_points, x_points[1:]))
        #         interval_list = [[x_points[i], x_points[i+1]] for i in range(len(x_points)-1)]

        intervals = [upper]
        intervals.extend([bisection_search(r_i, *i) for i in interval_list])
        intervals.append(lower)

        extermum_interval = [[intervals[i], intervals[i + 1]] for i in range(len(intervals) - 1)]

        extremums = [concave_max(r_i, *i) for i in extermum_interval]

        extremums[0] = mp.mpf(upper)
        extremums[-1] = mp.mpf(lower)

        errors = [abs(r_i(i)) for i in extremums]
        mean_error = numpy.mean(errors)

        if numpy.max([abs(error - mean_error) for error in errors]) < 0.000001 * mean_error:
            break

        x_points = extremums

    return [float(i) for i in numpy.polynomial.chebyshev.cheb2poly(coeffs)], float(mean_error)

def c_code_gen(data_type, name, poly_coeffs, comments = None):
    method_string = f'{data_type} {name} ({data_type} x)' + '{\n'

    if comments is not None:
        method_string += '\t// ' + str(comments) + ' \n\n'

    data_type_converter = '' if data_type == 'double' else 'f'

    method_string += '\n'.join([f'\tconst {data_type} a_{i} = {str(val) + data_type_converter};' for i, val in enumerate(poly_coeffs)])

    horner = 'return a_0+'
    for i in range(len(poly_coeffs)-2):
        horner += f'x*(a_{i+1} +'
    horner += f'x*a_{len(poly_coeffs)-1}' + ')'*(len(poly_coeffs)-2) + ';\n}'

    return method_string + '\n \t' + horner



from mpmath import *


class mpchebyshev:

    def __init__(self, f=cos, r=[1,2], d=5):
        mp.dps = 15
        mp.pretty = True
        self.poly, self.err = chebyfit(f, r, d, error=True)
        nprint(self.poly)
        nprint(self.err, 12)

    def __len__(self):
        return len(self.poly)

    def __getitem__(self, idx):
        return self.poly[idx]

    def __setitem__(self, idx, val):
        self.poly[idx] = val

    def predict(self, x):
        return nprint(polyval(self.poly, x), 12)



import unittest
from chiku.mpmath import mpdigits,mptaylor,mppade


class Test_MPMath(unittest.TestCase):
	def test_mpmath(self):
		d = mpdigits.mpdigits()
		one = d.get_d(1)

		def f(x):
			sqrt = d.sqrt()
			return sqrt((one + 2*x)/(one + x))

		self.t = mptaylor.mptaylor(f, 0, 6)

		self.p = mppade.mppade(self.t, 3, 3)
		self.x = 10

		self.assertAlmostEqual(self.p.predict(self.x), f(self.x), 2)

if __name__ == '__main__':
	unittest.main()



import unittest
from chiku import fourier


class Test_Fourier(unittest.TestCase):
	def test_fourier(self):
		def f(x):
			return x*x

		self.poly = fourier.fourier(f, -1, 5, 70)
		self.assertEqual(round(self.poly.predict(2)), f(2))

if __name__ == '__main__':
	unittest.main()



import math
from scipy.integrate import quad


class fourier:

    def __init__(self, f, a, b, k):
        self.L = b - a

        def integrand(x):
            return f(x)
        val, err = quad(integrand, a, b)
        self.a0 = (1/(2*self.L))*val

        self.ak = []
        for i in range(k):
            def integrand(x):
                return f(x)*math.cos(((i+1)*math.pi*x)/self.L)
            val, err = quad(integrand, a, b)
            self.ak.append((1/self.L)*val)

        self.bk = []
        for i in range(k):
            def integrand(x):
                return f(x)*math.sin(((i+1)*math.pi*x)/self.L)
            val, err = quad(integrand, a, b)
            self.bk.append((1/self.L)*val)


    def predict(self, x):
        res = self.a0

        for k in range(len(self.ak)):
            res += self.ak[k] * math.cos((math.pi*(k+1)*x)/self.L)

        for k in range(len(self.bk)):
            res += self.bk[k] * math.sin((math.pi*(k+1)*x)/self.L)

        return res



from scipy import linalg


class remez:

    def __init__(self, f, cheb, degree):

        if len(cheb)-2 != degree or len(cheb)%2 != 0:
            exit("order mismatch")

        matP = []
        matY = []
        ed = 1

        for x in cheb:
            matY.append(f(x))

            arrP = [1]
            for d in range(degree):
                arrP.append(x**(d+1))
            arrP.append((-1)**(1+ed))

            matP.append(arrP)
            ed += 1

        self.coeffs = linalg.solve(matP, matY)


    def __len__(self):
        return len(self.coeffs)

    def __getitem__(self, idx):
        return self.coeffs[idx]

    def __setitem__(self, idx, val):
        self.coeffs[idx] = val


    def predict(self, x):
        res = self.coeffs[-1]

        for i in range(len(self.coeffs)-1):
            res += self.coeffs[i] * (x**i)

        return res



import unittest
from chiku import chebyshev, remez
import numpy as np
import random

class Test_Remez(unittest.TestCase):
	def test_remez(self):
		def sigmoid(x):
			return 1/(1+np.exp(-x))

		self.cpoly = chebyshev.chebyshev(sigmoid, degree=6, frange=[-2,2])
		self.poly = remez.remez(sigmoid, self.cpoly.get_coeffs(), degree=4)

		self.rnum = random.random()
		self.assertAlmostEqual(self.poly.predict(self.rnum), sigmoid(self.rnum), 2)

if __name__ == '__main__':
	unittest.main()



from sympy import symbols, solve


class pade:

	def __init__(self, fcoeffs=[0.622459, 0.235004, -0.0287784, -0.0160595, 0.00436483, 0.00113017, -0.000542105], pd=4, qd=4):

		ta = fcoeffs
		tl = len(ta)
		aa = []
		al = pd
		ba = ["1"]
		bl = qd

		for a in range(al):
			aa.append("a"+str(a))

		for b in range(1,bl):
			ba.append("b"+str(b))

		tb = []

		for b in range(bl):
			temp = []
			for t in range(tl):
				temp.append(str(ta[t])+"*"+ba[b])
			tb.append(temp)

		eq = []

		for i in range(tl+bl-1):
			if i == 0:
				ia = [(0,0)]
			else:
				ia = list(set(ix))

			ix = []
			str_eq = ""

			for s in ia:
				ri,ci = s[0],s[1]
				str_eq += "(" + tb[ri][ci] + ") + "

				if ri+1 < bl:
					ix.append((ri+1,ci))
				if ci+1 < tl:
					ix.append((ri,ci+1))

			eq.append(str_eq[:-2])

		feq = []
		for e in range(len(eq)):
			if e < al:
				feq.append(eq[e] + "- " + aa[e])
			else:
				feq.append(eq[e])

		syma = []
		for a in aa:
			syma.append(a)
		for b in range(1,bl):
			syma.append(ba[b])

		solution = solve(feq[:(al+bl-1)], syma)
		print("Solution:", solution)

		self.pa = []
		self.qa = [1]

		for a in aa:
			self.pa.append(solution[symbols(a)])
		for b in range(1,bl):
			self.qa.append(solution[symbols(ba[b])])


	def get_p_coeffs(self):
		return self.pa

	def get_q_coeffs(self):
		return self.qa


	def predict(self, x):
		nom = 0
		den = 0

		for i in range(len(self.pa)):
			nom += self.pa[i]*(x**i)
		for i in range(len(self.qa)):
			den += self.qa[i]*(x**i)

		if den == 0:
			den = 1

		return nom/den



import math


class chebyshev:

	def __init__(self, funcx=math.log, degree=5, frange=[-1,1]):
		self.a = frange[0]
		self.b = frange[1]

		bma = 0.5 * (self.b - self.a)
		bpa = 0.5 * (self.b + self.a)
		f = [funcx(math.cos(math.pi * (k + 0.5) / degree) * bma + bpa) for k in range(degree)]
		fac = 2.0 / degree
		self.c = [fac * sum([f[k] * math.cos(math.pi * j * (k + 0.5) / degree)
			for k in range(degree)]) for j in range(degree)]


	def __len__(self):
		return len(self.c)

	def __getitem__(self, idx):
		return self.c[idx]

	def __setitem__(self, idx, val):
		self.c[idx] = val

	def get_coeffs(self):
		return self.c

	def print_coeffs(self):
		print(self.c)


	def predict(self, x):
		y = (2.0 * x - self.a - self.b) * (1.0 / (self.b - self.a))
		y2 = 2.0 * y
		(d, dd) = (self.c[-1], 0)             # Special case first step for efficiency
		for cj in self.c[-2:0:-1]:            # Clenshaw's recurrence
			(d, dd) = (y2 * d - dd + cj, d)

		return y * d - dd + 0.5 * self.c[0]   # Last step is different



from mpmath import *


class mpfourier:

    def __init__(self, f=lambda x: x**2 - 4*x + 1, r=[-1, 1.5], d=4):
        mp.dps = 15
        mp.pretty = True

        self.r = r
        self.cs = fourier(f, r, d)

        nprint(self.cs[0])
        nprint(self.cs[1])


    def predict(self, x):
        return fourierval(self.cs, self.r, x)
